@@ -834,3 +834,11 @@ void SignalSwitcher::restore_handle()
 #else /* other architectures - might not be entirely precise, and will be slower */
     size_t log2ceil(size_t x) {return (size_t)(ceill(log2l((long double) x)));}
 #endif
+
+#ifdef _FOR_PYTHON
+ModelOutputs deepcopy(const ModelOutputs &inp)
+{
+    ModelOutputs out = inp;
+    return out;
+}
+#endif
