@@ -52,6 +52,7 @@
 #include <stdlib.h>
 #include <stddef.h>
 #include <string.h>
+#include <stdint.h>
 #ifdef _OPENMP
     #include <omp.h>
 #endif
@@ -115,6 +116,8 @@ typedef enum SplitType {
 } SplitType;
 typedef enum ColTransf {NoTransf, Log, Exp} ColTransf; /* transformation to apply to numeric column */
 
+/* TODO: should write serializers for the model objects, but need to somehow deal with long double types */
+
 /*    
 *    1-d clusters that define homogeneous groups in which observations can be outliers.
 *    Note that these are associated to a tree and define one extra condition from what
@@ -126,7 +129,7 @@ typedef struct Cluster {
     size_t col_num = 0; /* numer of the column by which its being split, the target column is given by index of the cluster vector */
     SplitType split_type = Root;
     double split_point = HUGE_VAL; /* numerical */
-    std::vector<signed char> split_subset = std::vector<signed char>(); /* categorical */
+    std::vector<signed char> split_subset; /* categorical */
     int split_lev = INT_MAX;    /* ordinal */
     bool has_NA_branch = false; /* this is in order to determine the best outlier cluster when it fits under more than 1 */
 
@@ -139,7 +142,7 @@ typedef struct Cluster {
     double    display_lim_high = -HUGE_VAL;                /* numerical target column */
     double    display_mean = -HUGE_VAL;                    /* numerical target column */
     double    display_sd = -HUGE_VAL;                      /* numerical target column */
-    std::vector<signed char> subset_common = std::vector<signed char>(); /* categorical or ordinal target column (=0 is common) */
+    std::vector<signed char> subset_common; /* categorical or ordinal target column (=0 is common) */
     double    perc_in_subset = HUGE_VAL;                   /* categorical or ordinal target column */
     double    perc_next_most_comm = -HUGE_VAL;             /* categorical or ordinal target column */ /* TODO */
     int       categ_maj = -1;                              /* when using majority-criterion for categorical outliers */
@@ -257,21 +260,21 @@ typedef struct Cluster {
 typedef struct ClusterTree {
     size_t parent = 0;              /* index in a vector */
     SplitType parent_branch = Root; /* this tree follows this branch in the split given by its parent */
-    std::vector<size_t> clusters = std::vector<size_t>(); /* these clusters define additional splits */
+    std::vector<size_t> clusters; /* these clusters define additional splits */
 
     SplitType split_this_branch = Root;                        /* when using 'follow_all' */
-    std::vector<size_t> all_branches = std::vector<size_t>();  /* when using 'follow_all' */
+    std::vector<size_t> all_branches;  /* when using 'follow_all' */
 
     ColType   column_type = NoType;
     size_t    col_num = 0;
     double    split_point = HUGE_VAL;
-    std::vector<signed char> split_subset = std::vector<signed char>();
+    std::vector<signed char> split_subset;
     int split_lev = INT_MAX;
 
     size_t tree_NA = 0;    /* binary splits */
     size_t tree_left = 0;  /* binary splits */
     size_t tree_right = 0; /* binary splits */
-    std::vector<size_t> binary_branches = std::vector<size_t>(); /* multiple splits (single category or binarized categories) */
+    std::vector<size_t> binary_branches; /* multiple splits (single category or binarized categories) */
 
     ClusterTree(size_t parent, SplitType parent_branch)
     {
@@ -348,6 +351,8 @@ typedef struct ClusterTree {
 
 } ClusterTree;
 
+/* TODO: should separate the results from the actual model object */
+
 /* these are needed for prediction time, and are thus returned from the function that fits the model */
 typedef struct ModelOutputs {
     std::vector< std::vector<ClusterTree> > all_trees;  /* clusters in which observations can be outliers, required for prediction time */
@@ -382,11 +387,6 @@ typedef struct ModelOutputs {
         archive(
                 this->all_trees,
                 this->all_clusters,
-                this->outlier_scores_final,
-                this->outlier_clusters_final,
-                this->outlier_columns_final,
-                this->outlier_trees_final,
-                this->outlier_depth_final,
                 this->start_ix_cat_counts,
                 this->prop_categ,
                 this->col_transf,
@@ -433,7 +433,67 @@ bool fit_outliers_models(ModelOutputs &model_outputs,
                          size_t max_depth = 3, double max_perc_outliers = 0.01, size_t min_size_numeric = 25, size_t min_size_categ = 50,
                          double min_gain = 1e-2, bool gain_as_pct = false, bool follow_all = false, double z_norm = 2.67, double z_outlier = 8.0);
 
-typedef struct {
+class ExhaustedColumnTracker
+{
+public:
+    std::vector<bool> is_exhausted;
+    std::vector<size_t> col_indices;
+    std::vector<size_t> n_held;
+
+    void initialize(size_t ncols, size_t max_depth)
+    {
+        this->is_exhausted.assign(ncols, false);
+        this->n_held.clear();
+        this->n_held.reserve(max_depth+1);
+        this->col_indices.clear();
+        this->col_indices.reserve(ncols);
+    }
+
+    void push_branch()
+    {
+        this->n_held.push_back(0);
+    }
+
+    void push_col(size_t col)
+    {
+        this->is_exhausted[col] = true;
+        this->col_indices.push_back(col);
+        this->n_held.back() += 1;
+    }
+
+    void pop_branch()
+    {
+        size_t col;
+        while (this->n_held.back() > 0)
+        {
+            col = this->col_indices.back();
+            this->is_exhausted[col] = false;
+            this->col_indices.pop_back();
+            this->n_held.back() -= 1;
+        }
+
+        this->n_held.pop_back();
+    }
+};
+
+class ExhaustedColumnsLevel
+{
+public:
+    bool pop = false;
+    ExhaustedColumnTracker* tracker = nullptr;
+    ExhaustedColumnsLevel() = default;
+    void initialize(ExhaustedColumnTracker* tracker) {
+        this->pop = true;
+        this->tracker = tracker;
+        this->tracker->push_branch();
+    }
+    ~ExhaustedColumnsLevel() {
+        if (this->pop)
+            this->tracker->pop_branch();
+    }
+};
+
+struct Workspace {
     
     std::vector<size_t> ix_arr;           /* indices from the target column */
     size_t st;                            /* chunk of the indices to take for current function calls */
@@ -489,8 +549,8 @@ typedef struct {
     std::vector<size_t>      buffer_crosstab;        /* buffer arrays where to allocate values required by functions and not used outside them */
     std::vector<size_t>      buffer_cat_cnt;         /* buffer arrays where to allocate values required by functions and not used outside them */
     std::vector<size_t>      buffer_cat_sorted;      /* buffer arrays where to allocate values required by functions and not used outside them */
-    std::vector<signed char>        buffer_subset_categ;    /* buffer arrays where to allocate values required by functions and not used outside them */
-    std::vector<signed char>        buffer_subset_outlier;  /* buffer arrays where to allocate values required by functions and not used outside them */
+    std::vector<signed char> buffer_subset_categ;    /* buffer arrays where to allocate values required by functions and not used outside them */
+    std::vector<signed char> buffer_subset_outlier;  /* buffer arrays where to allocate values required by functions and not used outside them */
     std::vector<long double> buffer_sd;              /* used for a more numerically-stable two-pass gain calculation */
     
     bool drop_cluster;          /* for categorical and ordinal variables, not all clusters can flag observations as outliers, so those are not kept */
@@ -498,10 +558,13 @@ typedef struct {
     bool target_col_is_ord;     /* whether the target column is ordinal (rest is the same as in categoricals) */
     int  ncat_this;             /* number of categories in the target column */
 
-} Workspace;
+    ExhaustedColumnTracker exhausted_col_tracker;
+    bool has_zero_variance;
+
+};
 
 /* info holders to shorten function call arguments */
-typedef struct {
+struct ModelParams {
     bool    categ_as_bin;
     bool    ord_as_bin;
     bool    cat_bruteforce_subset;
@@ -518,16 +581,16 @@ typedef struct {
     double  z_outlier;
     double  z_tail;
     std::vector<long double> prop_small; /* this is not a parameter, but a shared array determined from the parameters and data */
-} ModelParams;
+};
 
 /* Note: the vectors here are filled within the function that fits the model, while the pointers are passed from outside */
-typedef struct {
+struct InputData {
     double  *restrict numeric_data;     size_t ncols_numeric;
     int     *restrict categorical_data; size_t ncols_categ;   int *restrict ncat;
     int     *restrict ordinal_data;     size_t ncols_ord;     int *restrict ncat_ord;
     size_t  nrows; size_t tot_cols; std::vector<char> has_NA; std::vector<char> skip_col; int max_categ;
     std::vector<size_t> cat_counts;
-} InputData;
+};
 
 
 void process_numeric_col(std::vector<Cluster> &cluster_root,
@@ -559,12 +622,12 @@ void recursive_split_categ(Workspace &workspace,
     (This is the module from which
      new data can be flagged as outliers)
 ********************************************/
-typedef struct {
+struct PredictionData {
     double  *restrict numeric_data;
     int     *restrict categorical_data;
     int     *restrict ordinal_data;
     size_t nrows;
-} PredictionData;
+};
 
 bool find_new_outliers(double *restrict numeric_data,
                        int    *restrict categorical_data,
@@ -582,19 +645,21 @@ bool check_is_outlier_in_tree(std::vector<size_t> &clusters_in_tree, size_t curr
 *********************************/
 #define SD_REG 1e-5 /* Regularization for standard deviation estimation */
 
-typedef struct {
+/* TODO: should make long doubles optional */
+
+struct NumericBranch {
     size_t      cnt;
     long double sum;
     long double sum_sq;
-} NumericBranch;
+};
 
-typedef struct {
+struct NumericSplit {
     NumericBranch NA_branch    = {0, 0, 0};
     NumericBranch left_branch  = {0, 0, 0};
     NumericBranch right_branch = {0, 0, 0};
-} NumericSplit;
+};
 
-typedef struct {
+struct CategSplit {
     size_t *restrict NA_branch;     /* array of counts of the target variable's categories */
     size_t *restrict left_branch;   /* array of counts of the target variable's categories */
     size_t *restrict right_branch;  /* array of counts of the target variable's categories */
@@ -603,7 +668,7 @@ typedef struct {
     size_t size_NA    = 0;
     size_t size_left  = 0;
     size_t size_right = 0;
-} CategSplit;
+};
 
 void subset_to_onehot(size_t ix_arr[], size_t n_true, size_t n_tot, signed char onehot[]);
 size_t move_zero_count_to_front(size_t *restrict cat_sorted, size_t *restrict cat_cnt, size_t ncat_x);
@@ -622,23 +687,25 @@ long double categ_gain_from_split(size_t *restrict ix_arr, int *restrict x, size
                                   size_t ncat, size_t *restrict buffer_cat_cnt, long double base_info);
 void split_numericx_numericy(size_t *restrict ix_arr, size_t st, size_t end, double *restrict x, double *restrict y,
                              long double sd_y, bool has_na, size_t min_size, bool take_mid, long double *restrict buffer_sd,
-                             long double *restrict gain, double *restrict split_point, size_t *restrict split_left, size_t *restrict split_NA);
+                             long double *restrict gain, double *restrict split_point, size_t *restrict split_left, size_t *restrict split_NA, bool *restrict has_zero_variance);
 void split_categx_numericy(size_t *restrict ix_arr, size_t st, size_t end, int *restrict x, double *restrict y, long double sd_y, double ymean,
                            bool x_is_ordinal, size_t ncat_x, size_t *restrict buffer_cat_cnt, long double *restrict buffer_cat_sum,
                            long double *restrict buffer_cat_sum_sq, size_t *restrict buffer_cat_sorted,
-                           bool has_na, size_t min_size, long double *gain, signed char *restrict split_subset, int *restrict split_point);
+                           bool has_na, size_t min_size, long double *gain, signed char *restrict split_subset, int *restrict split_point, bool *restrict has_zero_variance);
 void split_numericx_categy(size_t *restrict ix_arr, size_t st, size_t end, double *restrict x, int *restrict y,
                            size_t ncat_y, long double base_info, size_t *restrict buffer_cat_cnt,
                            bool has_na, size_t min_size, bool take_mid, long double *restrict gain, double *restrict split_point,
-                           size_t *restrict split_left, size_t *restrict split_NA);
+                           size_t *restrict split_left, size_t *restrict split_NA, bool *restrict has_zero_variance);
 void split_ordx_categy(size_t *restrict ix_arr, size_t st, size_t end, int *restrict x, int *restrict y,
                        size_t ncat_y, size_t ncat_x, long double base_info,
                        size_t *restrict buffer_cat_cnt, size_t *restrict buffer_crosstab, size_t *restrict buffer_ord_cnt,
-                       bool has_na, size_t min_size, long double *gain, int *split_point);
+                       bool has_na, size_t min_size, long double *gain, int *split_point,
+                       bool *restrict has_zero_variance);
 void split_categx_biny(size_t *restrict ix_arr, size_t st, size_t end, int *restrict x, int *restrict y,
                        size_t ncat_x, long double base_info,
                        size_t *restrict buffer_cat_cnt, size_t *restrict buffer_crosstab, size_t *restrict buffer_cat_sorted,
-                       bool has_na, size_t min_size, long double *gain, signed char *restrict split_subset);
+                       bool has_na, size_t min_size, long double *gain, signed char *restrict split_subset,
+                       bool *restrict has_zero_variance);
 void split_categx_categy_separate(size_t *restrict ix_arr, size_t st, size_t end, int *restrict x, int *restrict y,
                                   size_t ncat_x, size_t ncat_y, long double base_info,
                                   size_t *restrict buffer_cat_cnt, size_t *restrict buffer_crosstab,
@@ -646,7 +713,8 @@ void split_categx_categy_separate(size_t *restrict ix_arr, size_t st, size_t end
 void split_categx_categy_subset(size_t *restrict ix_arr, size_t st, size_t end, int *restrict x, int *restrict y,
                                 size_t ncat_x, size_t ncat_y, long double base_info,
                                 size_t *restrict buffer_cat_cnt, size_t *restrict buffer_crosstab, size_t *restrict buffer_split,
-                                bool has_na, size_t min_size, long double *gain, signed char *restrict split_subset);
+                                bool has_na, size_t min_size, long double *gain, signed char *restrict split_subset,
+                                bool *restrict has_zero_variance);
 
 
 
@@ -788,3 +856,7 @@ void check_interrupt_switch(SignalSwitcher &ss);
 bool cy_check_interrupt_switch();
 void cy_tick_off_interrupt_switch();
 #endif
+size_t log2ceil(size_t v);
+#ifdef _FOR_PYTHON
+ModelOutputs deepcopy(const ModelOutputs &inp);
+#endif
