@@ -226,6 +226,11 @@ bool follow_tree(ModelOutputs &model_outputs, PredictionData &prediction_data, s
                                                      true : found_outliers;
                                 break;
                             }
+
+                            default:
+                            {
+                                assert(0);
+                            }
                         }
                         break;
                     }
@@ -279,6 +284,11 @@ bool follow_tree(ModelOutputs &model_outputs, PredictionData &prediction_data, s
                                                      true : found_outliers;
                                 break;
                             }
+
+                            default:
+                            {
+                                assert(0);
+                            }
                         }
                         break;
                     }
@@ -332,10 +342,16 @@ bool follow_tree(ModelOutputs &model_outputs, PredictionData &prediction_data, s
                                                      true : found_outliers;
                                 break;
                             }
+
+                            default:
+                            {
+                                assert(0);
+                            }
                         }
                         break;
                     }
 
+                    default: {}
                 }
             }
         }
@@ -548,6 +564,11 @@ bool check_is_outlier_in_tree(std::vector<size_t> &clusters_in_tree, size_t curr
                             if (!isnan(num_val_other) && num_val_other > model_outputs.all_clusters[col][cl].split_point) flag_this_cluster = true;
                             break;
                         }
+
+                        default:
+                        {
+                            assert(0);
+                        }
                     }
                     break;
                 }
@@ -587,6 +608,11 @@ bool check_is_outlier_in_tree(std::vector<size_t> &clusters_in_tree, size_t curr
                             break;
                         }
 
+                        default:
+                        {
+                            assert(0);
+                        }
+
                         /* Note: type 'SingleCateg' is only used temporarily, later gets converted to 'Equal' */
                     }
                     break;
@@ -626,6 +652,11 @@ bool check_is_outlier_in_tree(std::vector<size_t> &clusters_in_tree, size_t curr
                             if (cat_val_other >=0 && cat_val_other != model_outputs.all_clusters[col][cl].split_lev) flag_this_cluster = true;
                             break;
                         }
+
+                        default:
+                        {
+                            assert(0);
+                        }
                     }
                     break;
                 }
